/* !
 * FunLazy v3.0.0
 * Copyright (C) 2020-present, ZG
 * Released under the MIT license.
 */
!( ( global, factory ) => {
    typeof exports === "object" && typeof module !== "undefined" ? module.exports = factory() :
    typeof define === "function" && define.amd ? define( factory ) :
    ( global = global || self, global.FunLazy = factory() );
} )( typeof window !== "undefined" ? window : this, () => {

    "use strict";

    // 缓存 DOM
    const document = window.document;
    const docElem = document.documentElement;
    const $head = document.head;

    // 默认配置
    const _placeholder = "data:img/jpg;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAMAAAAoyzS7AAAAA1BMVEXd3d3u346CAAAADUlEQVR42gECAP3/AAAAAgABUyucMAAAAABJRU5ErkJggg==";
    const defaults = {
        container: "body",
        effect: "show",
        placeholder: _placeholder,
        threshold: 0,
        width: null,
        height: null,
        axis: "y",
        eventType: "",
        autoCheckChange: false,
        useErrorImagePlaceholder: false,
        strictLazyMode: true,
        beforeLazy: () => {},
        onSuccess: () => {},
        onError: () => {}
    };

    // 淡入效果
    const duration = 400;
    const fadeInCSS = `.funlazy-fadeIn{animation:funlazy-fadeIn ${ duration }ms ease 0s 1 normal forwards running;}@keyframes funlazy-fadeIn{from{opacity:0}to{opacity:1}}`;

    // 动态添加 style
    const styleId = "funlazy-fade-in-style";
    const $style = document.getElementById( styleId );
    if ( $style ) {
        $head.removeChild( $style );
    }
    $head.insertAdjacentHTML( "afterbegin", `<style id="${ styleId }">${ fadeInCSS }</style>` );

    // 获取样式
    const getCSS = ( elem, prop ) => window.getComputedStyle( elem, null ).getPropertyValue( prop );

    // 设置图片
    const setImage = ( el, src, isImg ) => {
        if ( isImg ) {
            el.src = src;
        } else {
            el.style.backgroundImage = `url(${ src })`;
        }
    }

    // 判断目标元素的祖先元素中是否有隐藏的元素
    let hasHidden = false;
    let parent = null;
    const parentsHasHidden = elem => {
        if ( document.body.contains( elem ) ) {
            if ( elem.parentNode.nodeName.toLowerCase() !== "body" ) {
                parent = elem.parentNode;
                if ( getCSS( parent, "display" ) === "none" || getCSS( parent, "visibility" ) === "hidden" ) {
                    hasHidden = true;
                } else {
                    parentsHasHidden( parent );
                }
            }
        }
        return hasHidden;
    }

    // FunLazy 函数
    const FunLazy = options => {

        // 合并参数
        const opt = ( ( options, defaults ) => {
            const obj = {};
            for ( const key in defaults ) {
                obj[ key ] = options[ key ] === undefined ? defaults[ key ] : options[ key ];
            }
            return obj;
        } )( options || {}, defaults );

        // 检测事件是否为函数形式
        const beforeLazyIsFunc = typeof opt.beforeLazy === "function",
            successIsFunc = typeof opt.onSuccess === "function",
            errorIsFunc = typeof opt.onError === "function";

        // 目标元素
        let $container;
        if ( opt.container === window || opt.container === document || opt.container === "html" ) {
            opt.container = "body";
        }
        if ( typeof opt.container === "string" ) {
            $container = document.querySelector( opt.container );
        }
        if ( !$container ) {
            return;
        }

        // 需要执行懒加载操作的元素必须设置了 data-funlazy 属性
        const $funlazy = $container.querySelectorAll( "[data-funlazy]" );
        if ( !$funlazy.length ) {
            return;
        }

        Array.from( $funlazy ).forEach( $this => {

            // 图片真实地址
            let lazy = $this.getAttribute( "data-funlazy" ).trim(); 

            // 含有以下任一条件的元素将不进行懒加载操作
            // - 没有设置有效的 data-funlazy 属性
            // - 已经成功进行过懒加载操作的元素
            if ( !lazy || $this.funlazyLoaded ) {
                return;
            }

            // 当设置了 strictLazyMode: true 时
            // 含有以下任一条件的元素将不进行懒加载操作
            // - 此元素本身处于隐藏状态
            // - 此元素含有处于隐藏状态的祖先元素
            // - 此元素本身 opacity: 0
            if ( opt.strictLazyMode ) {
                if ( 
                    getCSS( $this, "display" ) === "none" || 
                    getCSS( $this, "visibility" ) === "hidden" || 
                    parentsHasHidden( $this ) ||
                    +getCSS( $this, "opacity" ) === 0
                ) {
                    return;
                }
            }

            // 是否为 <img> 元素
            const isImg = $this.nodeName.toLowerCase() === "img";

            // 添加占位图片
            if ( typeof opt.placeholder === "string" ) {
                setImage( $this, opt.placeholder, isImg );
            }

            // 通过脚本设置宽高
            // 也可以通过 css 或属性设置
            if ( opt.width ) {
                $this.style.width = typeof opt.width === "string" ? opt.width : `${ opt.width }px`;
            }
            if ( opt.height ) {
                $this.style.height = typeof opt.height === "string" ? opt.height : `${ opt.height }px`;
            }

            // 加载图片
            const load = ( $target, callback ) => {

                // 如果此元素已经完成了懒加载操作
                // 则直接执行回调
                if ( $target.funlazyLoaded ) {
                    callback && callback( true );
                    return;
                }

                // 创建一个 <img> 执行加载操作
                const image = new Image();
                image.onload = () => {

                    // 将真实图片地址赋给元素
                    setImage( $target, lazy, isImg );

                    // fadeIn 淡入效果
                    if ( opt.effect === "fadeIn" && "classList" in docElem ) {
                        $target.classList.add( "funlazy-fadeIn" );
                        let timer = window.setTimeout( () => {
                            window.clearTimeout( timer );
                            timer = null;
                            $target.classList.remove( "funlazy-fadeIn" );
                        }, duration );
                    }

                    // 执行 success 回调
                    if ( !$target.funlazyLoaded && successIsFunc ) {
                        opt.onSuccess( $target, lazy );
                    }

                    // 添加成功标识
                    $target.funlazyLoaded = true;

                    callback && callback( true );
                }
                image.onerror = () => {
                        
                    // 执行 error 回调
                    if ( !$target.funlazyFailed && errorIsFunc ) {
                        opt.onError( $target, lazy );
                    }

                    // 设置错误图片占位符
                    const usePlaceholder = opt.useErrorImagePlaceholder;
                    if ( usePlaceholder ) {

                        // 无论是使用默认占位图片还是自定义图片
                        // 默都先使用默认占位图片进行占位
                        $target.src = _placeholder;

                        // 使用自定义图片
                        if ( typeof usePlaceholder === "string" && usePlaceholder.trim() ) {

                            // 需要检查图片的可用性
                            const img = new Image();
                            img.src = usePlaceholder.trim();
                            img.onload = () => {
                                $target.src = usePlaceholder.trim();
                            }
                        }
                    }

                    // 添加失败标识
                    $target.funlazyFailed = true;

                    callback && callback( false );
                }

                // 加载图片之前执行的事件
                if ( beforeLazyIsFunc ) {
                    lazy = opt.beforeLazy( lazy ) || lazy;
                }

                image.src = lazy;
            }

            // 监听
            const listener = () => {
                if ( !$this.funlazyLoaded ) {
                    const threshold = typeof opt.threshold === "number" && opt.threshold > 0 ? opt.threshold : 0;
                    const io = new IntersectionObserver( entries => {
                        entries.forEach( item => {
                            if ( item.target && item.isIntersecting ) {
                                load( item.target, bool => {
                                    if ( bool === true ) {
                                        io.unobserve( item.target );
                                    }
                                } );
                            }
                        } )
                    }, {
                        root: opt.container === "body" ? null : $container,
                        threshold: [ 0 ],
                        rootMargin: `0px ${ opt.axis === "x" ? threshold : 0 }px ${ opt.axis === "y" ? threshold : 0 }px 0px`
                    } );
                    io.observe( $this );
                }
            }
            
            // 指定了鼠标事件类型
            if ( opt.eventType.match( /(click|dblclick|mouseover)/ ) ) {
                document.addEventListener( opt.eventType, event => {
                    if ( event.target === $this && !$this.funlazyLoaded && !$this.funlazyFailed ) {
                        load( $this );
                    }
                } );
            } else {
                listener();
            }
        } )

        // 实时检测目标元素内含有 data-funlazy 属性的元素的变化情况
        // 可以对新增的懒加载元素进行自动解析
        if ( opt.autoCheckChange ) {
            if ( supportMutationObserver ) {
                new MutationObserver( () => {
                    FunLazy( opt );
                } ).observe( $container, {
                    childList: true,
                    subtree: true
                } );
            }
        }

        return opt;     
    }

    return FunLazy;

} );